//@ts-nocheck
// https://www.joezimjs.com/javascript/javascript-design-patterns-decorator/

/* ---
	First we’ll create the base object that will be a Car with no cool features.
	This also sets up the interface that the decorators will use.
																																			--- */

var Car = function() {
	console.log( 'Assemble: build frame, add core parts' );
}

/* ---
	The decorators will also need to implement these functions
	to comply with Car's interface.
																												--- */


Car.prototype = {
	start: function() {
		console.log( 'The engine starts with roar!' );
	},
	drive: function() {
		console.log( 'Away we go!' );
	},
	getPrice: function() {
		return 11000.00;
	}
}

/* -->
	Now we’ll create the decorator “class” that each of the decorator will inherit from.
	You’ll notice that each of the functions simply pass the call on to the Car that
	they’re wrapping -- In this case the only functions that will be overridden are
	assemble and getPrice.

	Pass the Car (or CarDecorator) instance as the parameter to the CarDecorator's
	base object, in order to be able to add features to it, like so:
						                                                               <-- */

// - Establishes a base object for `CarDecorator`
var CarDecorator = function( car ) {
	/** @type {CarDecorator} */
	this.car = car;
}


/* CarDecorator implements the same interface as Car itself */
CarDecorator.prototype = {
	start: function() {
		this.car.start();
	},
	drive: function() {
		this.car.drive();
	},
	getPrice: function() {
		return this.car.getPrice();
	}
}

/* -->
	Next, we create a decorator object for each feature, and override the
	parent’s functions whenever we want to expand or replace the functionality.
																																			<-- */

var PowerLocksDecorator = function( car ) {
	/* JavaScript's way of calling a parent class' constructor */
	CarDecorator.call( this, car );

	// Do things ...
}
PowerLocksDecorator.prototype = new CarDecorator();

PowerLocksDecorator.prototype.drive = function() {
	/*
	 You can either do this ...							  	<--	*/

	this.car.drive();
	/* ... or, you can call the parent's drive function:
		CarDecorator.prototype.drive.call(this);
																							<--*/
}

var PowerWindowsDecorator = function( car ) {
	CarDecorator.call( this, car );

	console.log( 'Assemble: add power windows' );
}

PowerWindowsDecorator.prototype = new CarDecorator();

var ACDecorator = function( car ) {
	CarDecorator.call( this, car );

	console.log( 'Assemble: add A/C unit' );
}

ACDecorator.prototype = new CarDecorator();

ACDecorator.prototype.start = function() {
	this.car.start();

	console.log( 'The cool air starts blowing.' );
}